/* OpenVAS-Client
 * $Id$
 * Description: SSH Key management.
 *
 * Authors:
 * Felix Wolfsteller <felix.wolfsteller@intevation.de>
 *
 * Copyright:
 * Copyright (C) 2008 Intevation GmbH
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2,
 * or, at your option, any later version as published by the Free
 * Software Foundation
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * In addition, as a special exception, you have
 * permission to link the code of this program with the OpenSSL
 * library (or with modified versions of OpenSSL that use the same
 * license as OpenSSL), and distribute linked combinations including
 * the two. You must obey the GNU General Public License in all
 * respects for all of the code used other than OpenSSL. If you
 * modify this file, you may extend this exception to your version
 * of the file, but you are not obligated to do so. If you do not
 * wish to do so, delete this exception statement from your version.
 */

#include "openvas_ssh_login.h"
#include "nessus_i18n.h"
#include "error_dlg.h"
#include <glib/gstdio.h>

#define KEY_SSHLOGIN_USERNAME "username"
#define KEY_SSHLOGIN_PUBKEY_FILE "pubkey_file"
#define KEY_SSHLOGIN_PRIVKEY_FILE "privkey_file"
#define KEY_SSHLOGIN_COMMENT "comment"
#define KEY_SSHLOGIN_PASSPHRASE "passphrase"

/**
 * Ensures the existance of a directory by creating it if it does not exist.
 * ATTENTION: can only create "one more sublevel"!
 * @param directory Path of directory to check or create.
 * @return TRUE if directory exists or was successfully created, FALSE otherwise.
 */
static gboolean ensure_dir(char* directory)
{
  //Should be moved together with file functions defined in context.h/c
  // to an own module e.g. file_check.c
  if(check_is_dir(directory) == 1)
    return TRUE;
  // Glib >= 2.8 defines g_mkdir_with_parents
  if(g_mkdir(directory, 0700) == 0)
    return TRUE;
  else
    return FALSE;
}

/**
 * Forks and creates a key for local checks by calling
 * "openssl pkcs8 -topk8 -v2 des3 -in filepath -passin pass:passphrase -out
 *          filepath.p8 -passout pass:passphrase"
 * The key is placed in a Context- specific directory.
 * @param name Name of the key (part of filename) (will be freed).
 * @param passphrase The passphrase for the key (will be freed), must be longer
 *                   than 4 characters (+nul).
 * @return TRUE if successfull, FALSE otherwise.
 */
static
gboolean openvas_ssh_privkey_create(char* pubkey_file, char* privkey_file,
                                    char* passphrase_pub, char* passphrase_priv)
{
  gchar* stdout = NULL;
  gchar* stderr = NULL;
  GError* err = NULL;
  gint exit_status;
  char* dir;
  
  /* Sanity-check essential parameters */
  if(!passphrase_pub || !passphrase_priv)
  {
    show_error(_("Error creating private key file:\nPlease provide all information."));
    return FALSE;
  }
  
  /* Sanity check files */
  if(check_exists(pubkey_file) != 1)
  {
    show_error(_("Error creating private key file:\nPublic key %s not found."), pubkey_file);
    return FALSE;
  }
  if(check_exists(privkey_file) != 0 )
  {
    show_error(_("Error creating private key file:\nFile already exists."));
    return FALSE;
  }
  dir = g_path_get_dirname(privkey_file);
  if(ensure_dir(dir) != TRUE)
  {
    show_error(_("Error creating private key file:\nfolder %s not accessible."), dir);
    efree(&dir);
    return FALSE;
  }
  efree(&dir);

  /* Fire openssl */
  const char* command = g_strconcat("openssl pkcs8 -topk8 -v2 des3 -in ", pubkey_file,
                                    " -passin pass:", passphrase_pub, " -out ",
                                    privkey_file, " -passout pass:",
                                    passphrase_priv, NULL);
  
  if(g_spawn_command_line_sync(command, &stdout, &stderr, &exit_status, &err) == FALSE
          || exit_status != 0 )
  {
    show_error(_("Error creating private key file.\nFor further information consult your shell."));
    printf("Error creating private key file.");
    printf("\tSpawned openssl process returned with %d.\n", exit_status);
    printf("\t\t stdout: %s", stdout);
    printf("\t\t stderr: %s", stderr);
    return FALSE;
  }

  return TRUE;
}

/**
 * Forks and creates a key for local checks by calling
 * "ssh-keygen -t rsa -f filepath -C comment -P passhprase -q"
 * A directory will be created if it does not exist.
 * @param comment Comment to use (will be freed).
 * @param passphrase The passphrase for the key (will be freed), must be longer
 *                   than 4 characters (+nul).
 * @param filepath Path to file of public key (trailing .pub will be stripped).
 * @return TRUE if successfull, FALSE otherwise.
 */
static
gboolean openvas_ssh_pubkey_create(char* comment, char* passphrase,
                                   char* filepath)
{
  gchar* stdout = NULL;
  gchar* stderr = NULL;
  GError* err = NULL;
  gint exit_status;
  char* dir;
  char* file_pubstripped;
  
  /* Sanity-check essential parameters */
  if( !comment || comment[0] == '\0' )
  {
    show_error(_("Error creating public key file:\ncomment has to be set."));
    return FALSE;
  }
  if(!passphrase || strlen(passphrase) < 5)
  {
    show_error(_("Error creating public key file:\npassword must be longer than 4 characters."));
    return FALSE;
  }
  /* Sanity check files */
  dir = g_path_get_dirname(filepath);
  if(ensure_dir(dir) != TRUE)
  {
    show_error(_("Error creating public key file:\n%s is not accessable."), filepath);
    efree(&dir);
    return FALSE;
  }
  efree(&dir);
  if(check_exists(filepath) == 1)
  {
    show_error(_("Error creating public key file:\n%s already exists."), filepath);
    return FALSE;
  }
  
  // Strip ".pub" of filename, if any.
  if(g_str_has_suffix(filepath, ".pub") == TRUE)
  {
    file_pubstripped = emalloc(strlen(filepath) - strlen(".pub") +1);
    strncpy(file_pubstripped, filepath, strlen(filepath) - strlen(".pub"));
  }
  else
    file_pubstripped = g_strdup(filepath);
  
  /* Fire ssh-keygen */
  const char* command = g_strconcat("ssh-keygen -t rsa -f ", filepath, " -C ",
                                    comment, " -P ", passphrase, NULL);
  efree(&file_pubstripped);

  if(g_spawn_command_line_sync(command, &stdout, &stderr, &exit_status, &err) == FALSE
     || exit_status != 0 )
  {
    show_error(_("Error creating public key file.\nFor further information consult your shell."));
    printf("Error creating public key file.\n");
    printf("\tSpawned key-gen process returned with %d.\n", exit_status);
    printf("\t\t stdout: %s", stdout);
    printf("\t\t stderr: %s", stderr);
    return FALSE;
  }
  return TRUE;
}


/**
 * Initializes a openvas_ssh_login.
 * Key and Info files have to be created separately.
 * @return A fresh openvas_ssh_login.
 */
openvas_ssh_login* openvas_ssh_login_new(char* name, char* pubkey_file, char* privkey_file,
                                         char* passphrase, char* comment,
                                         char* uname)
{
  openvas_ssh_login* loginfo = emalloc(sizeof(openvas_ssh_login));
  loginfo->name = name;
  loginfo->username = uname;
  loginfo->public_key_path = pubkey_file;
  loginfo->private_key_path = privkey_file;
  loginfo->ssh_key_passphrase = passphrase;
  loginfo->comment = comment;

  return loginfo;
}


/**
 * Frees data associated with a openvas_ssh_login.
 * @param loginfo openvas_ssh_login to free.
 */
void openvas_ssh_login_free(openvas_ssh_login* loginfo)
{
  if(loginfo == NULL)
    return;
  if(loginfo->name)
    efree(&loginfo->name);
  if(loginfo->username)
    efree(&loginfo->username);
  if(loginfo->public_key_path)
    efree(&loginfo->public_key_path);
  if(loginfo->private_key_path)
    efree(&loginfo->private_key_path);
  if(loginfo->ssh_key_passphrase)
    efree(&loginfo->ssh_key_passphrase);
  if(loginfo->comment)
    efree(&loginfo->comment);
  efree(&loginfo);
}


/**
 * Creates the public and private key files.
 * @param loginfo.
 * @return TRUE if things went good, FALSE if things went bad.
 */
gboolean openvas_ssh_login_create(openvas_ssh_login* loginfo)
{
  /* Create pubkey */
  gboolean success = openvas_ssh_pubkey_create(loginfo->comment, 
                                               loginfo->ssh_key_passphrase,
                                               loginfo->public_key_path);
  
  /* Eventually report failure */
  if(success == FALSE)
    return FALSE;
  
  /* Create private key */
  success = openvas_ssh_privkey_create(loginfo->public_key_path,
                                       loginfo->private_key_path,
                                       loginfo->ssh_key_passphrase,
                                       loginfo->ssh_key_passphrase);
  return success;
}

/**
 * Creates a string to be sent to the server as value for a SSH_LOGIN
 * plugin preference.
 * It follows the pattern:
 * username|userpass|pubkeyfilepath|privkeyfilepath|passphrase .
 * @param loginfo openvas_ssh_login that will be used to assemble the string.
 * @return Freshly created string or NULL if loginfo == NULL.
 */
char*
openvas_ssh_login_prefstring(openvas_ssh_login* loginfo)
{
  if(loginfo != NULL)
    return g_strjoin("|", loginfo->username, loginfo->public_key_path, 
                     loginfo->private_key_path, loginfo->ssh_key_passphrase,
                     NULL);
  else return NULL;
}
    

// ---------------- File store functions ------------------
 
 
/**
 * Callback for a g_hashtable_for_each. Adds entries to a GKeyFile.
 */
static void add_sshlogin_to_file(char* name, openvas_ssh_login* loginfo, 
                                 GKeyFile* key_file)
{
  if(name == NULL || key_file == NULL || loginfo == NULL)
    return;

  g_key_file_set_string(key_file, loginfo->name, KEY_SSHLOGIN_USERNAME, 
                        loginfo->username);
  g_key_file_set_string(key_file, loginfo->name, KEY_SSHLOGIN_PUBKEY_FILE, 
                        loginfo->public_key_path);
  g_key_file_set_string(key_file, loginfo->name, KEY_SSHLOGIN_PRIVKEY_FILE, 
                        loginfo->private_key_path);
  g_key_file_set_string(key_file, loginfo->name, KEY_SSHLOGIN_COMMENT, 
                        loginfo->comment);
  g_key_file_set_string(key_file, loginfo->name, KEY_SSHLOGIN_PASSPHRASE, 
                        loginfo->ssh_key_passphrase);
}


/**
 * Writes information of all ssh logins found in a hashtable into a file.
 * To load the information again, openvas_ssh_login_file_read can be used.
 * @param ssh_logins Hashtable with pointers to openvas_ssh_login s as values.
 * @param filename Path to file to wtite to.
 * @return TRUE if file was written (success), FALSE if an error occured.
 */
gboolean openvas_ssh_login_file_write(GHashTable* ssh_logins, char* filename)
{
  GKeyFile* key_file = g_key_file_new();
  gchar* keyfile_data;
  gsize data_length;
  GError* err = NULL;
  int fd;
  
  g_key_file_set_comment(key_file, NULL, NULL, 
                         "This file was generated by OpenVAS and shall not be edited manually.",
                         &err);
  if (err != NULL)
  {
    show_error(_("Error adding comment to key file: %s"), err->message);
    g_error_free(err);
    g_key_file_free(key_file);
    return FALSE;
  }

  // Add all ssh logins to GKeyFile.
  if(ssh_logins != NULL)
  {
    g_hash_table_foreach(ssh_logins, (GHFunc) add_sshlogin_to_file, key_file);    
  } // (else file content is comment only)
  
  // Write GKeyFile to filesystem.
  fd = open(filename, O_RDWR|O_CREAT|O_TRUNC, 0600);
  if(!fd)
  {
    show_error(_("Error accessing ssh info file."));
    g_key_file_free(key_file);
    return FALSE;
  }

  keyfile_data = g_key_file_to_data(key_file, &data_length, &err);
  if(err != NULL)
  {
    show_error(_("Error exporting ssh info file: %s"), err->message);
    g_error_free(err);
    g_key_file_free(key_file);
    return FALSE;
  }

  write(fd, keyfile_data, data_length);
  close(fd);

  g_key_file_free(key_file);

  return TRUE;
}


/**
 * Reads a ssh_login file and returns GHashTable with
 * the names as keys and pointers to openvas_ssh_logins as values.
 * openvas_ssh_logins are checked before being added to the hashtable:
 * if the public and private key files do not exist, the openvas_ssh_login is
 * not added.
 * @param filename File to read from.
 * @return GHashTable, keys are names of openvas_ssh_logins, who are values.
 */
GHashTable* openvas_ssh_login_file_read(char* filename)
{
  gchar** names;
  gsize length;
  GKeyFile* key_file = g_key_file_new();
  GError* err        = NULL;
  GHashTable* loginfos   = g_hash_table_new_full(g_str_hash, g_str_equal, 
      NULL, (GDestroyNotify) openvas_ssh_login_free);

  g_key_file_load_from_file(key_file, filename, G_KEY_FILE_NONE, &err);

  if(err != NULL)
  {
    // No file found? Thats ok, return empty hashtable.
    if(err->code == G_KEY_FILE_ERROR_NOT_FOUND || err->code == G_FILE_ERROR_NOENT)
    {
      g_key_file_free(key_file);
      return loginfos;
    }
      
    g_hash_table_destroy(loginfos);
    show_error(_("Error loading sshlogin store %s: %s"), filename,
               err->message);
    g_key_file_free(key_file);
    return NULL;
  }

  names = g_key_file_get_groups(key_file, &length);

  // Read ssh login information from file and add entry to hashtable.
  int i = 0;
  for(i = 0; i < length; i++)
  {
    if(names[i] == NULL || names[i] == '\0')
      continue;
    // Init a openvas_ssh_login
    char* name = names[i];
    char* username = g_key_file_get_string(key_file, names[i], 
                                           KEY_SSHLOGIN_USERNAME, &err);
    char* pubkey   = g_key_file_get_string(key_file, names[i], 
                                           KEY_SSHLOGIN_PUBKEY_FILE, &err);
    char* privkey  = g_key_file_get_string(key_file, names[i], 
                                           KEY_SSHLOGIN_PRIVKEY_FILE, &err);
    char* comment  = g_key_file_get_string(key_file, names[i], 
                                           KEY_SSHLOGIN_COMMENT, &err);
    char* passphrase = g_key_file_get_string(key_file, names[i], 
                                             KEY_SSHLOGIN_PASSPHRASE, &err);
    
    openvas_ssh_login* loginfo = openvas_ssh_login_new(name, pubkey, privkey,
                                  passphrase, comment, username);

    // Discard if error or files do not exist
    if(err != NULL || check_exists(pubkey) == 0 || check_exists(privkey) == 0)
    {
      openvas_ssh_login_free(loginfo);
    }
    else
    {
      // Add to hash table otherwise
      g_hash_table_insert(loginfos, loginfo->name, loginfo);
    }
  }

  g_key_file_free(key_file);

  return loginfos;
}
